@@ -285,6 +285,14 @@ class TestS3SigV4Auth(BaseTestWithFixedDate):
         self.request.method = 'PUT'
         self.request.url = 'https://s3.eu-central-1.amazonaws.com/'
 
+        self.client_config = mock.Mock()
+        self.s3_config = {}
+        self.client_config.s3 = self.s3_config
+
+        self.request.context = {
+            'client_config': self.client_config
+        }
+
     def test_resign_with_content_hash(self):
         self.auth.add_auth(self.request)
         original_auth = self.request.headers['Authorization']
@@ -319,6 +327,78 @@ class TestS3SigV4Auth(BaseTestWithFixedDate):
         cqs = self.auth.canonical_query_string(request)
         self.assertEqual('marker=%C3%A4%C3%B6%C3%BC-01.txt&prefix=', cqs)
 
+    def _test_blacklist_header(self, header, value):
+        request = AWSRequest()
+        request.url = 'https://s3.amazonaws.com/bucket/foo'
+        request.method = 'PUT'
+        request.headers[header] = value
+        credentials = botocore.credentials.Credentials('access_key',
+                                                       'secret_key')
+        auth = botocore.auth.S3SigV4Auth(credentials, 's3', 'us-east-1')
+        auth.add_auth(request)
+        self.assertNotIn(header, request.headers['Authorization'])
+
+    def test_blacklist_expect_headers(self):
+        self._test_blacklist_header('expect', '100-continue')
+
+    def test_blacklist_headers(self):
+        self._test_blacklist_header('user-agent', 'botocore/1.4.11')
+
+    def test_context_sets_signing_region(self):
+        original_signing_region = 'eu-central-1'
+        new_signing_region = 'us-west-2'
+        self.auth.add_auth(self.request)
+        auth = self.request.headers['Authorization']
+        self.assertIn(original_signing_region, auth)
+        self.assertNotIn(new_signing_region, auth)
+
+        self.request.context = {'signing': {'region': new_signing_region}}
+        self.auth.add_auth(self.request)
+        auth = self.request.headers['Authorization']
+        self.assertIn(new_signing_region, auth)
+        self.assertNotIn(original_signing_region, auth)
+
+    def test_uses_sha256_if_config_value_is_true(self):
+        self.client_config.s3['payload_signing_enabled'] = True
+        self.auth.add_auth(self.request)
+        sha_header = self.request.headers['X-Amz-Content-SHA256']
+        self.assertNotEqual(sha_header, 'UNSIGNED-PAYLOAD')
+
+    def test_does_not_use_sha256_if_config_value_is_false(self):
+        self.client_config.s3['payload_signing_enabled'] = False
+        self.auth.add_auth(self.request)
+        sha_header = self.request.headers['X-Amz-Content-SHA256']
+        self.assertEqual(sha_header, 'UNSIGNED-PAYLOAD')
+
+    def test_uses_sha256_if_md5_unset(self):
+        self.request.context['has_streaming_input'] = True
+        self.auth.add_auth(self.request)
+        sha_header = self.request.headers['X-Amz-Content-SHA256']
+        self.assertNotEqual(sha_header, 'UNSIGNED-PAYLOAD')
+
+    def test_uses_sha256_if_not_https(self):
+        self.request.context['has_streaming_input'] = True
+        self.request.headers.add_header('Content-MD5', 'foo')
+        self.request.url = 'http://s3.amazonaws.com/bucket'
+        self.auth.add_auth(self.request)
+        sha_header = self.request.headers['X-Amz-Content-SHA256']
+        self.assertNotEqual(sha_header, 'UNSIGNED-PAYLOAD')
+
+    def test_uses_sha256_if_not_streaming_upload(self):
+        self.request.context['has_streaming_input'] = False
+        self.request.headers.add_header('Content-MD5', 'foo')
+        self.request.url = 'http://s3.amazonaws.com/bucket'
+        self.auth.add_auth(self.request)
+        sha_header = self.request.headers['X-Amz-Content-SHA256']
+        self.assertNotEqual(sha_header, 'UNSIGNED-PAYLOAD')
+
+    def test_does_not_use_sha256_if_md5_set(self):
+        self.request.context['has_streaming_input'] = True
+        self.request.headers.add_header('Content-MD5', 'foo')
+        self.auth.add_auth(self.request)
+        sha_header = self.request.headers['X-Amz-Content-SHA256']
+        self.assertEqual(sha_header, 'UNSIGNED-PAYLOAD')
+
 
 class TestSigV4(unittest.TestCase):
     def setUp(self):
@@ -668,6 +748,7 @@ class BaseS3PresignPostTest(unittest.TestCase):
         self.request.context['s3-presign-post-fields'] = self.fields
         self.request.context['s3-presign-post-policy'] = self.policy
 
+
 class TestS3SigV2Post(BaseS3PresignPostTest):
     def setUp(self):
         super(TestS3SigV2Post, self).setUp()
@@ -719,6 +800,7 @@ class TestS3SigV2Post(BaseS3PresignPostTest):
         self.assertEqual(result_policy['conditions'], [])
         self.assertIn('signature', result_fields)
 
+
 class TestS3SigV4Post(BaseS3PresignPostTest):
     def setUp(self):
         super(TestS3SigV4Post, self).setUp()
