@@ -20,10 +20,12 @@ import botocore.auth
 from botocore.compat import six, urlparse, parse_qs
 
 from botocore.credentials import Credentials
+from botocore.credentials import ReadOnlyCredentials
 from botocore.exceptions import NoRegionError, UnknownSignatureVersionError
 from botocore.exceptions import UnknownClientMethodError, ParamValidationError
 from botocore.exceptions import UnsupportedSignatureVersionError
 from botocore.signers import RequestSigner, S3PostPresigner, CloudFrontSigner
+from botocore.config import Config
 
 from tests import unittest
 
@@ -36,6 +38,7 @@ class BaseSignerTest(unittest.TestCase):
         self.signer = RequestSigner(
             'service_name', 'region_name', 'signing_name',
             'v4', self.credentials, self.emitter)
+        self.fixed_credentials = self.credentials.get_frozen_credentials()
 
 
 class TestSigner(BaseSignerTest):
@@ -65,33 +68,10 @@ class TestSigner(BaseSignerTest):
 
             self.assertEqual(auth, auth_cls.return_value)
             auth_cls.assert_called_with(
-                credentials=self.credentials, service_name='service_name',
+                credentials=self.fixed_credentials,
+                service_name='service_name',
                 region_name='region_name')
 
-    def test_get_auth_cached(self):
-        def side_effect(*args, **kwargs):
-            return mock.Mock()
-        auth_cls = mock.Mock(side_effect=side_effect)
-        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS,
-                             {'v4': auth_cls}):
-            auth1 = self.signer.get_auth('service_name', 'region_name')
-            auth2 = self.signer.get_auth('service_name', 'region_name')
-
-        self.assertEqual(auth1, auth2)
-
-    def test_get_auth_cached_expires(self):
-        def side_effect(*args, **kwargs):
-            return mock.Mock()
-        auth_cls = mock.Mock(side_effect=side_effect)
-        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS,
-                             {'v4': auth_cls}):
-            auth1 = self.signer.get_auth('service_name', 'region_name',
-                                         expires=60)
-            auth2 = self.signer.get_auth('service_name', 'region_name',
-                                         expires=90)
-
-        self.assertNotEqual(auth1, auth2)
-
     def test_get_auth_signature_override(self):
         auth_cls = mock.Mock()
         with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS,
@@ -101,7 +81,8 @@ class TestSigner(BaseSignerTest):
 
             self.assertEqual(auth, auth_cls.return_value)
             auth_cls.assert_called_with(
-                credentials=self.credentials, service_name='service_name',
+                credentials=self.fixed_credentials,
+                service_name='service_name',
                 region_name='region_name')
 
     def test_get_auth_bad_override(self):
@@ -131,8 +112,8 @@ class TestSigner(BaseSignerTest):
                              {'custom': auth}):
             self.signer.sign('operation_name', request)
 
-        auth.assert_called_with(credentials=self.credentials)
-        auth.return_value.add_auth.assert_called_with(request=request)
+        auth.assert_called_with(credentials=self.fixed_credentials)
+        auth.return_value.add_auth.assert_called_with(request)
 
     def test_emits_before_sign(self):
         request = mock.Mock()
@@ -160,6 +141,62 @@ class TestSigner(BaseSignerTest):
 
         auth.assert_not_called()
 
+    def test_generate_url_emits_choose_signer(self):
+        request_dict = {
+            'headers': {},
+            'url': 'https://foo.com',
+            'body': b'',
+            'url_path': '/',
+            'method': 'GET',
+            'context': {}
+        }
+
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS,
+                             {'v4': mock.Mock()}):
+            self.signer.generate_presigned_url(request_dict, 'operation_name')
+
+        self.emitter.emit_until_response.assert_called_with(
+            'choose-signer.service_name.operation_name',
+            signing_name='signing_name', region_name='region_name',
+            signature_version='v4-query')
+
+    def test_generate_url_choose_signer_override(self):
+        request_dict = {
+            'headers': {},
+            'url': 'https://foo.com',
+            'body': b'',
+            'url_path': '/',
+            'method': 'GET',
+            'context': {}
+        }
+        auth = mock.Mock()
+        auth.REQUIRES_REGION = False
+        self.emitter.emit_until_response.return_value = (None, 'custom')
+
+        auth_types_map = {'custom': mock.Mock(), 'custom-query': auth}
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types_map):
+            self.signer.generate_presigned_url(request_dict, 'operation_name')
+
+        auth.assert_called_with(credentials=self.fixed_credentials,
+                                expires=3600)
+
+    def test_generate_url_unsigned(self):
+        request_dict = {
+            'headers': {},
+            'url': 'https://foo.com',
+            'body': b'',
+            'url_path': '/',
+            'method': 'GET',
+            'context': {}
+        }
+        self.emitter.emit_until_response.return_value = (
+            None, botocore.UNSIGNED)
+
+        url = self.signer.generate_presigned_url(
+            request_dict, 'operation_name')
+
+        self.assertEqual(url, 'https://foo.com')
+
     def test_generate_presigned_url(self):
         auth = mock.Mock()
         auth.REQUIRES_REGION = True
@@ -169,13 +206,15 @@ class TestSigner(BaseSignerTest):
             'url': 'https://foo.com',
             'body': b'',
             'url_path': '/',
-            'method': 'GET'
+            'method': 'GET',
+            'context': {}
         }
         with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS,
                              {'v4-query': auth}):
-            presigned_url = self.signer.generate_presigned_url(request_dict)
+            presigned_url = self.signer.generate_presigned_url(
+                request_dict, operation_name='operation_name')
         auth.assert_called_with(
-            credentials=self.credentials, region_name='region_name',
+            credentials=self.fixed_credentials, region_name='region_name',
             service_name='signing_name', expires=3600)
         self.assertEqual(presigned_url, 'https://foo.com')
 
@@ -188,14 +227,16 @@ class TestSigner(BaseSignerTest):
             'url': 'https://foo.com',
             'body': b'',
             'url_path': '/',
-            'method': 'GET'
+            'method': 'GET',
+            'context': {}
         }
         with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS,
                              {'v4-query': auth}):
             presigned_url = self.signer.generate_presigned_url(
-                request_dict, region_name='us-west-2')
+                request_dict, operation_name='operation_name',
+                region_name='us-west-2')
         auth.assert_called_with(
-            credentials=self.credentials, region_name='us-west-2',
+            credentials=self.fixed_credentials, region_name='us-west-2',
             service_name='signing_name', expires=3600)
         self.assertEqual(presigned_url, 'https://foo.com')
 
@@ -208,48 +249,195 @@ class TestSigner(BaseSignerTest):
             'url': 'https://foo.com',
             'body': b'',
             'url_path': '/',
-            'method': 'GET'
+            'method': 'GET',
+            'context': {}
         }
         with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS,
                              {'v4-query': auth}):
             presigned_url = self.signer.generate_presigned_url(
-                request_dict, expires_in=900)
+                request_dict, operation_name='operation_name', expires_in=900)
         auth.assert_called_with(
-            credentials=self.credentials, region_name='region_name',
+            credentials=self.fixed_credentials,
+            region_name='region_name',
             expires=900, service_name='signing_name')
         self.assertEqual(presigned_url, 'https://foo.com')
 
-    def test_generate_presigned_url_fixes_s3_host(self):
-        self.signer = RequestSigner(
-            'service_name', 'region_name', 'signing_name',
-            's3', self.credentials, self.emitter)
-
-        auth = mock.Mock()
-        auth.REQUIRES_REGION = True
-
+    def test_presigned_url_throws_unsupported_signature_error(self):
         request_dict = {
             'headers': {},
             'url': 'https://s3.amazonaws.com/mybucket/myobject',
             'body': b'',
             'url_path': '/',
-            'method': 'GET'
+            'method': 'GET',
+            'context': {}
         }
-        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS,
-                             {'s3-query': auth}):
-            presigned_url = self.signer.generate_presigned_url(
-                request_dict, expires_in=900)
-        auth.assert_called_with(
-            credentials=self.credentials, region_name='region_name',
-            expires=900, service_name='signing_name')
-        self.assertEqual(presigned_url,
-                         'https://mybucket.s3.amazonaws.com/myobject')
-
-    def test_presigned_url_throws_unsupported_signature_error(self):
         self.signer = RequestSigner(
             'service_name', 'region_name', 'signing_name',
             'foo', self.credentials, self.emitter)
         with self.assertRaises(UnsupportedSignatureVersionError):
-            self.signer.generate_presigned_url({})
+            self.signer.generate_presigned_url(
+                request_dict, operation_name='foo')
+
+    def test_signer_with_refreshable_credentials_gets_credential_set(self):
+        class FakeCredentials(Credentials):
+            def get_frozen_credentials(self):
+                return ReadOnlyCredentials('foo', 'bar', 'baz')
+        self.credentials = FakeCredentials('a', 'b', 'c')
+
+        self.signer = RequestSigner(
+            'service_name', 'region_name', 'signing_name',
+            'v4', self.credentials, self.emitter)
+
+        auth_cls = mock.Mock()
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS,
+                             {'v4': auth_cls}):
+            auth = self.signer.get_auth('service_name', 'region_name')
+            self.assertEqual(auth, auth_cls.return_value)
+            # Note we're called with 'foo', 'bar', 'baz', and *not*
+            # 'a', 'b', 'c'.
+            auth_cls.assert_called_with(
+                credentials=ReadOnlyCredentials('foo', 'bar', 'baz'),
+                service_name='service_name',
+                region_name='region_name')
+
+    def test_no_credentials_case_is_forwarded_to_signer(self):
+        # If no credentials are given to the RequestSigner, we should
+        # forward that fact on to the Auth class and let them handle
+        # the error (which they already do).
+        self.credentials = None
+        self.signer = RequestSigner(
+            'service_name', 'region_name', 'signing_name',
+            'v4', self.credentials, self.emitter)
+        auth_cls = mock.Mock()
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS,
+                             {'v4': auth_cls}):
+            auth = self.signer.get_auth_instance(
+                'service_name', 'region_name', 'v4')
+            auth_cls.assert_called_with(
+                service_name='service_name',
+                region_name='region_name',
+                credentials=None,
+            )
+
+    def test_sign_with_signing_type_standard(self):
+        auth = mock.Mock()
+        post_auth = mock.Mock()
+        query_auth = mock.Mock()
+        request = mock.Mock()
+        auth_types = {
+            'v4-presign-post': post_auth,
+            'v4-query': query_auth,
+            'v4': auth
+        }
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
+            self.signer.sign('operation_name', request,
+                             signing_type='standard')
+        self.assertFalse(post_auth.called)
+        self.assertFalse(query_auth.called)
+        auth.assert_called_with(
+            credentials=ReadOnlyCredentials('key', 'secret', None),
+            service_name='signing_name',
+            region_name='region_name'
+        )
+
+    def test_sign_with_signing_type_presign_url(self):
+        auth = mock.Mock()
+        post_auth = mock.Mock()
+        query_auth = mock.Mock()
+        request = mock.Mock()
+        auth_types = {
+            'v4-presign-post': post_auth,
+            'v4-query': query_auth,
+            'v4': auth
+        }
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
+            self.signer.sign('operation_name', request,
+                             signing_type='presign-url')
+        self.assertFalse(post_auth.called)
+        self.assertFalse(auth.called)
+        query_auth.assert_called_with(
+            credentials=ReadOnlyCredentials('key', 'secret', None),
+            service_name='signing_name',
+            region_name='region_name'
+        )
+
+    def test_sign_with_signing_type_presign_post(self):
+        auth = mock.Mock()
+        post_auth = mock.Mock()
+        query_auth = mock.Mock()
+        request = mock.Mock()
+        auth_types = {
+            'v4-presign-post': post_auth,
+            'v4-query': query_auth,
+            'v4': auth
+        }
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
+            self.signer.sign('operation_name', request,
+                             signing_type='presign-post')
+        self.assertFalse(auth.called)
+        self.assertFalse(query_auth.called)
+        post_auth.assert_called_with(
+            credentials=ReadOnlyCredentials('key', 'secret', None),
+            service_name='signing_name',
+            region_name='region_name'
+        )
+
+    def test_sign_with_region_name(self):
+        request = mock.Mock()
+        auth = mock.Mock()
+        auth_types = {
+            'v4': auth
+        }
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
+            self.signer.sign('operation_name', request, region_name='foo')
+        auth.assert_called_with(
+            credentials=ReadOnlyCredentials('key', 'secret', None),
+            service_name='signing_name',
+            region_name='foo'
+        )
+
+    def test_sign_with_expires_in(self):
+        request = mock.Mock()
+        auth = mock.Mock()
+        auth_types = {
+            'v4': auth
+        }
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
+            self.signer.sign('operation_name', request, expires_in=2)
+        auth.assert_called_with(
+            credentials=ReadOnlyCredentials('key', 'secret', None),
+            service_name='signing_name',
+            region_name='region_name',
+            expires=2
+        )
+
+    def test_unknown_signer_raises_unknown_on_standard(self):
+        request = mock.Mock()
+        auth = mock.Mock()
+        auth_types = {
+            'v4': auth
+        }
+        self.emitter.emit_until_response.return_value = (None, 'custom')
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
+            with self.assertRaises(UnknownSignatureVersionError):
+                self.signer.sign('operation_name', request,
+                                 signing_type='standard')
+
+    def test_unknown_signer_raises_unsupported_when_not_standard(self):
+        request = mock.Mock()
+        auth = mock.Mock()
+        auth_types = {
+            'v4': auth
+        }
+        self.emitter.emit_until_response.return_value = (None, 'custom')
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
+            with self.assertRaises(UnsupportedSignatureVersionError):
+                self.signer.sign('operation_name', request,
+                                 signing_type='presign-url')
+
+            with self.assertRaises(UnsupportedSignatureVersionError):
+                self.signer.sign('operation_name', request,
+                                 signing_type='presign-post')
 
 
 class TestCloudfrontSigner(unittest.TestCase):
@@ -274,11 +462,11 @@ class TestCloudfrontSigner(unittest.TestCase):
             ip_address='12.34.56.78/9')
         expected = {
             "Statement": [{
-                "Resource":"foo",
-                "Condition":{
-                    "DateGreaterThan":{"AWS:EpochTime":1448928000},
-                    "DateLessThan":{"AWS:EpochTime":1451606400},
-                    "IpAddress":{"AWS:SourceIp":"12.34.56.78/9"}
+                "Resource": "foo",
+                "Condition": {
+                    "DateGreaterThan": {"AWS:EpochTime": 1448928000},
+                    "DateLessThan": {"AWS:EpochTime": 1451606400},
+                    "IpAddress": {"AWS:SourceIp": "12.34.56.78/9"}
                 },
             }]
         }
@@ -317,10 +505,10 @@ class TestCloudfrontSigner(unittest.TestCase):
         expected = (
             'http://test.com/index.html?foo=bar'
             '&Policy=eyJTdGF0ZW1lbnQiOlt7IlJlc291cmNlIjoiZm9vIiwiQ29uZ'
-                'Gl0aW9uIjp7IkRhdGVMZXNzVGhhbiI6eyJBV1M6RXBvY2hUaW1lIj'
-                'oxNDUxNjA2NDAwfSwiSXBBZGRyZXNzIjp7IkFXUzpTb3VyY2VJcCI'
-                '6IjEyLjM0LjU2Ljc4LzkifSwiRGF0ZUdyZWF0ZXJUaGFuIjp7IkFX'
-                'UzpFcG9jaFRpbWUiOjE0NDg5MjgwMDB9fX1dfQ__'
+            'Gl0aW9uIjp7IkRhdGVMZXNzVGhhbiI6eyJBV1M6RXBvY2hUaW1lIj'
+            'oxNDUxNjA2NDAwfSwiSXBBZGRyZXNzIjp7IkFXUzpTb3VyY2VJcCI'
+            '6IjEyLjM0LjU2Ljc4LzkifSwiRGF0ZUdyZWF0ZXJUaGFuIjp7IkFX'
+            'UzpFcG9jaFRpbWUiOjE0NDg5MjgwMDB9fX1dfQ__'
             '&Signature=c2lnbmVk&Key-Pair-Id=MY_KEY_ID')
         self.assertEqualUrl(signed_url, expected)
 
@@ -337,12 +525,14 @@ class TestS3PostPresigner(BaseSignerTest):
             'url': 'https://s3.amazonaws.com/mybucket',
             'body': b'',
             'url_path': '/',
-            'method': 'POST'
+            'method': 'POST',
+            'context': {}
         }
         self.auth = mock.Mock()
         self.auth.REQUIRES_REGION = True
         self.add_auth = mock.Mock()
         self.auth.return_value.add_auth = self.add_auth
+        self.fixed_credentials = self.credentials.get_frozen_credentials()
 
         self.datetime_patch = mock.patch('botocore.signers.datetime')
         self.datetime_mock = self.datetime_patch.start()
@@ -361,7 +551,7 @@ class TestS3PostPresigner(BaseSignerTest):
             post_form_args = self.signer.generate_presigned_post(
                 self.request_dict)
         self.auth.assert_called_with(
-            credentials=self.credentials, region_name='region_name',
+            credentials=self.fixed_credentials, region_name='region_name',
             service_name='signing_name')
         self.assertEqual(self.add_auth.call_count, 1)
         ref_request = self.add_auth.call_args[0][0]
@@ -373,6 +563,63 @@ class TestS3PostPresigner(BaseSignerTest):
                          'https://s3.amazonaws.com/mybucket')
         self.assertEqual(post_form_args['fields'], {})
 
+    def test_generate_presigned_post_emits_choose_signer(self):
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS,
+                             {'s3v4-presign-post': self.auth}):
+            self.signer.generate_presigned_post(self.request_dict)
+        self.emitter.emit_until_response.assert_called_with(
+            'choose-signer.service_name.PutObject',
+            signing_name='signing_name', region_name='region_name',
+            signature_version='s3v4-presign-post')
+
+    def test_generate_presigned_post_choose_signer_override(self):
+        auth = mock.Mock()
+        self.emitter.emit_until_response.return_value = (None, 'custom')
+        auth_types = {
+            's3v4-presign-post': self.auth,
+            'custom-presign-post': auth
+        }
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
+            self.signer.generate_presigned_post(self.request_dict)
+        auth.assert_called_with(
+            credentials=self.fixed_credentials, region_name='region_name',
+            service_name='signing_name')
+
+    def test_generate_presigne_post_choose_signer_override_known(self):
+        auth = mock.Mock()
+        self.emitter.emit_until_response.return_value = (
+            None, 's3v4-presign-post')
+        auth_types = {
+            's3v4-presign-post': self.auth,
+            'custom-presign-post': auth
+        }
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
+            self.signer.generate_presigned_post(self.request_dict)
+        self.auth.assert_called_with(
+            credentials=self.fixed_credentials, region_name='region_name',
+            service_name='signing_name')
+
+    def test_generate_presigned_post_bad_signer_raises_error(self):
+        auth = mock.Mock()
+        self.emitter.emit_until_response.return_value = (None, 's3-query')
+        auth_types = {
+            's3v4-presign-post': self.auth,
+            's3-query': auth
+        }
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS, auth_types):
+            with self.assertRaises(UnsupportedSignatureVersionError):
+                self.signer.generate_presigned_post(self.request_dict)
+
+    def test_generate_unsigned_post(self):
+        self.emitter.emit_until_response.return_value = (
+            None, botocore.UNSIGNED)
+        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS,
+                             {'s3v4-presign-post': self.auth}):
+            post_form_args = self.signer.generate_presigned_post(
+                self.request_dict)
+        expected = {'fields': {}, 'url': 'https://s3.amazonaws.com/mybucket'}
+        self.assertEqual(post_form_args, expected)
+
     def test_generate_presigned_post_with_conditions(self):
         conditions = [
             {'bucket': 'mybucket'},
@@ -383,7 +630,7 @@ class TestS3PostPresigner(BaseSignerTest):
             self.signer.generate_presigned_post(
                 self.request_dict, conditions=conditions)
         self.auth.assert_called_with(
-            credentials=self.credentials, region_name='region_name',
+            credentials=self.fixed_credentials, region_name='region_name',
             service_name='signing_name')
         self.assertEqual(self.add_auth.call_count, 1)
         ref_request = self.add_auth.call_args[0][0]
@@ -396,32 +643,24 @@ class TestS3PostPresigner(BaseSignerTest):
             self.signer.generate_presigned_post(
                 self.request_dict, region_name='foo')
         self.auth.assert_called_with(
-            credentials=self.credentials, region_name='foo',
+            credentials=self.fixed_credentials, region_name='foo',
             service_name='signing_name')
 
-    def test_generate_presigned_post_fixes_s3_host(self):
-        self.request_signer = RequestSigner(
-            'service_name', 'region_name', 'signing_name',
-            's3', self.credentials, self.emitter)
-        self.signer = S3PostPresigner(self.request_signer)
-
-        with mock.patch.dict(botocore.auth.AUTH_TYPE_MAPS,
-                             {'s3-presign-post': self.auth}):
-            post_form_args = self.signer.generate_presigned_post(
-                self.request_dict)
-        self.auth.assert_called_with(
-            credentials=self.credentials, region_name='region_name',
-            service_name='signing_name')
-        self.assertEqual(post_form_args['url'],
-                         'https://mybucket.s3.amazonaws.com/')
-
     def test_presigned_post_throws_unsupported_signature_error(self):
+        request_dict = {
+            'headers': {},
+            'url': 'https://s3.amazonaws.com/mybucket/myobject',
+            'body': b'',
+            'url_path': '/',
+            'method': 'POST',
+            'context': {}
+        }
         self.request_signer = RequestSigner(
             'service_name', 'region_name', 'signing_name',
             'foo', self.credentials, self.emitter)
         self.signer = S3PostPresigner(self.request_signer)
         with self.assertRaises(UnsupportedSignatureVersionError):
-            self.signer.generate_presigned_post({})
+            self.signer.generate_presigned_post(request_dict)
 
 
 class TestGenerateUrl(unittest.TestCase):
@@ -448,9 +687,11 @@ class TestGenerateUrl(unittest.TestCase):
             'headers': {},
             'query_string': {},
             'url_path': u'/mybucket/mykey',
-            'method': u'GET'}
+            'method': u'GET',
+            'context': {}}
         self.generate_url_mock.assert_called_with(
-            request_dict=ref_request_dict, expires_in=3600)
+            request_dict=ref_request_dict, expires_in=3600,
+            operation_name='GetObject')
 
     def test_generate_presigned_url_with_query_string(self):
         disposition = 'attachment; filename="download.jpg"'
@@ -467,9 +708,11 @@ class TestGenerateUrl(unittest.TestCase):
             'headers': {},
             'query_string': {u'response-content-disposition': disposition},
             'url_path': u'/mybucket/mykey',
-            'method': u'GET'}
+            'method': u'GET',
+            'context': {}}
         self.generate_url_mock.assert_called_with(
-            request_dict=ref_request_dict, expires_in=3600)
+            request_dict=ref_request_dict, expires_in=3600,
+            operation_name='GetObject')
 
     def test_generate_presigned_url_unknown_method_name(self):
         with self.assertRaises(UnknownClientMethodError):
@@ -489,9 +732,11 @@ class TestGenerateUrl(unittest.TestCase):
             'headers': {},
             'query_string': {},
             'url_path': u'/mybucket/mykey',
-            'method': u'GET'}
+            'method': u'GET',
+            'context': {}}
         self.generate_url_mock.assert_called_with(
-            request_dict=ref_request_dict, expires_in=20)
+            request_dict=ref_request_dict, expires_in=20,
+            operation_name='GetObject')
 
     def test_generate_presigned_url_override_http_method(self):
         self.client.generate_presigned_url(
@@ -503,9 +748,11 @@ class TestGenerateUrl(unittest.TestCase):
             'headers': {},
             'query_string': {},
             'url_path': u'/mybucket/mykey',
-            'method': u'PUT'}
+            'method': u'PUT',
+            'context': {}}
         self.generate_url_mock.assert_called_with(
-            request_dict=ref_request_dict, expires_in=3600)
+            request_dict=ref_request_dict, expires_in=3600,
+            operation_name='GetObject')
 
 
 class TestGeneratePresignedPost(unittest.TestCase):
